﻿using System;
using System.Text;

namespace DarwinDecriptor
{
    /// <summary>
    /// Class that convert <see cref="System.Guid"/> to Darwin Descriptor GUID and vise verse.
    /// </summary>
    public class GuidConverter
    {
        /// <summary>
        /// Table for decode base85 string
        /// </summary>
        private static readonly uint[] CharToOffsetMapping =
            {
                0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
                0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
                0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0xff, 0xff,
                0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0xff,
                0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0xff, 0xff,
                0xff, 0x16, 0xff, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
                0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b,
                0x2c, 0x2d, 0x2e, 0x2f, 0x30, 0x31, 0x32, 0x33, 0xff, 0x34, 0x35, 0x36,
                0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40, 0x41, 0x42,
                0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e,
                0x4f, 0x50, 0x51, 0x52, 0xff, 0x53, 0x54, 0xff,
            };

        private static readonly string table_enc85 = "!$%&'()*+,-.0123456789=?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[]^_`abcdefghijklmnopqrstuvwxyz{}~";

        /// <summary>
        /// Convert Darwin Descriptor GUID string in to <see cref="System.Guid"/> representation.
        /// </summary>
        /// <param name="encodedGuid">
        /// Darwin Descriptor GUID.
        /// </param>
        /// <returns>
        /// The decoded <see cref="System.Guid"/>.
        /// </returns>
        /// <exception cref="ArgumentNullException">
        /// <paramref name="encodedGuid"/> cannot be null.
        /// </exception>
        /// <exception cref="ArgumentException">
        /// <paramref name="encodedGuid"/> cannot be empty.
        /// </exception>
        public Guid Decode(string encodedGuid)
        {
            if (encodedGuid == null)
            {
                throw new ArgumentNullException("encodedGuid");
            }

            if (encodedGuid == string.Empty)
            {
                throw new ArgumentException("Input string cannot be empty", "encodedGuid");
            }

            if (encodedGuid.Length != 20)
            {
                throw new ArgumentException("Input string should be 20 chars long", "encodedGuid");
            }

            ulong total = 0;
            uint multiplier = 1;
            var guidParts = new string[4];

            for (var currentSymbolPosition = 0; currentSymbolPosition < 20; currentSymbolPosition++)
            {
                if (this.IsNotDarwinDescriptorSymbol(encodedGuid[currentSymbolPosition]))
                {
                    throw new ArgumentException("Input string has wrong format", "encodedGuid");
                }

                if (this.IsNewOctet(currentSymbolPosition))
                {
                    total = 0;
                    multiplier = 1;
                }
                
                total += CharToOffsetMapping[encodedGuid[currentSymbolPosition]] * multiplier;
                
                if (this.IsOctetEnded(currentSymbolPosition))
                {
                    guidParts[currentSymbolPosition / 5] = total.ToString("X8");
                }

                multiplier *= 85;
            }

            return this.BuildGuid(guidParts);
        }

        /// <summary>
        /// Convert <see cref="System.Guid"/> in to Darwin Descriptor GUID string.
        /// </summary>
        /// <param name="inputGuid">
        /// The <see cref="System.Guid"/> value to convert.
        /// </param>
        /// <returns>
        /// The Darwin Descriptor GUID string.
        /// </returns>
        public string Encode(Guid inputGuid)
        {
            var result = new StringBuilder(20);
            var guidParts = this.GetGuidParts(inputGuid);

            for (var currentGuidPartIndex = 0; currentGuidPartIndex < 4; currentGuidPartIndex++)
            {
                uint offset = guidParts[currentGuidPartIndex];
                for (int currentGuidSymbolIndex = 0; currentGuidSymbolIndex < 5; currentGuidSymbolIndex++)
                {
                    result.Append(this.GetDarwinDescriptorSymbol(offset));
                    offset = offset / 85;
                }
            }

            return result.ToString();
        }

        private char GetDarwinDescriptorSymbol(uint offset)
        {
            return table_enc85[(int)(offset % 85)];
        }

        private uint[] GetGuidParts(Guid inputGuid)
        {
            var guidString = inputGuid.ToString("N");

            var guidParts = new[]
                                {
                                    this.HexToUInt(guidString.Substring(0, 8)),
                                    this.HexToUInt(guidString.Substring(8, 8).ReverseString(4)),
                                    this.HexToUInt(guidString.Substring(16, 8).ReverseString(2)),
                                    this.HexToUInt(guidString.Substring(24, 8).ReverseString(2))
                                };
            return guidParts;
        }

        private Guid BuildGuid(string[] guidParts)
        {
            var guidBuilder = new StringBuilder(32);
            guidBuilder.Append(guidParts[0]);
            guidBuilder.Append(guidParts[1].ReverseString(4));
            guidBuilder.Append(guidParts[2].ReverseString(2));
            guidBuilder.Append(guidParts[3].ReverseString(2));

            return Guid.ParseExact(guidBuilder.ToString(), "N");
        }

        private bool IsNotDarwinDescriptorSymbol(char symbol)
        {
            return symbol >= 0x80 || CharToOffsetMapping[symbol] == 0xff;
        }

        private bool IsNewOctet(int position)
        {
            return (position % 5) == 0;
        }

        private bool IsOctetEnded(int position)
        {
            return (position % 5) == 4;
        }

        private uint HexToUInt(string hexValue)
        {
            return uint.Parse(hexValue, System.Globalization.NumberStyles.HexNumber);
        }
    }
}
